"""
Realm-related models for Solar Network Python SDK.

This module contains data classes for realms and realm members.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Dict, Any


@dataclass
class Realm:
    """Realm information."""
    id: str
    slug: str
    name: str = ""
    description: str = ""
    verified_as: Optional[str] = None
    verified_at: Optional[datetime] = None
    is_community: bool = False
    is_public: bool = False
    picture: Optional[CloudFile] = None
    background: Optional[CloudFile] = None
    account_id: str = ""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class RealmMember:
    """Realm member information."""
    realm_id: str = ""
    realm: Optional[Realm] = None
    account_id: str = ""
    account: Optional[Account] = None
    role: int = 0
    joined_at: Optional[datetime] = None
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    status: Optional[AccountStatus] = None


"""
Data models for Solar Network Python SDK.

This module contains all the data classes and enums used throughout the SDK.
"""

# Account models
from .accounts import (
    Account, AccountProfile, AccountStatus, AccountBadge, ContactMethod,
    Notification, VerificationMark, AuthDevice, AuthDeviceWithSession,
    ExperienceRecord, SocialCreditRecord, FriendOverviewItem,
)

# Auth models
from .auth import (
    AuthChallenge, AuthSession, AppToken, GeoIpLocation, AuthFactor,
    AccountConnection,
)

# Activity models
from .activity import (
    NotableDay, TimelineEvent, CheckInResult, FortuneTip, EventCalendarEntry,
    PresenceActivity,
)

# ActivityPub models
from .activitypub import (
    ActivityPubInstance, ActivityPubUser, ActivityPubActor,
    ActivityPubFollowResponse, ActorStatusResponse,
)

# Chat models
from .chat import (
    ChatRoom, ChatMessage, ChatReaction, ChatMember, ChatSummary,
    ChatOnlineAccount, ChatOnlineStatus, MessageSyncResponse,
    ChatRealtimeJoinResponse, CallParticipant, RealtimeCall,
)

# Drive models
from .drive import (
    FileReplica, CloudFileObject, CloudFile, CloudFileIndex,
    FilePool, CloudFolder, FileListItem, DriveTask, UniversalFile,
    UniversalFileType,
)

# Live models
from .live import (
    LiveStream, LiveStreamStatus, LiveStreamType, LiveStreamVisibility,
)

# Post models
from .posts import (
    Post, Publisher, PublisherStats, PublisherSubscription, ReactInfo,
    PostEmbedView, PostEmbedViewRenderer, PostAward, PostReaction,
    PostFeaturedRecord, PostWithStats, PollWithStats, Poll, PollQuestion,
    PollOption, PollQuestionType, PollAnswer, PostTag, CategorySubscription,
    PostCategory, ScrappedLink, Heatmap, HeatmapItem,
)

# Realm models
from .realms import (
    Realm, RealmMember,
)

# Wallet models
from .wallets import (
    Wallet, WalletStats, WalletPocket, Transaction, WalletSubscription,
    WalletSubscriptionRef, WalletOrder, WalletGift, WalletFund,
    WalletFundRecipient, LotteryTicket, LotteryRecord,
)

__all__ = [
    # Account models
    'Account', 'AccountProfile', 'AccountStatus', 'AccountBadge',
    'ContactMethod', 'Notification', 'VerificationMark', 'AuthDevice',
    'AuthDeviceWithSession', 'ExperienceRecord', 'SocialCreditRecord',
    'FriendOverviewItem',
    
    # Auth models
    'AuthChallenge', 'AuthSession', 'AppToken', 'GeoIpLocation',
    'AuthFactor', 'AccountConnection',
    
    # Activity models
    'NotableDay', 'TimelineEvent', 'CheckInResult', 'FortuneTip',
    'EventCalendarEntry', 'PresenceActivity',
    
    # ActivityPub models
    'ActivityPubInstance', 'ActivityPubUser', 'ActivityPubActor',
    'ActivityPubFollowResponse', 'ActorStatusResponse',
    
    # Chat models
    'ChatRoom', 'ChatMessage', 'ChatReaction', 'ChatMember', 'ChatSummary',
    'ChatOnlineAccount', 'ChatOnlineStatus', 'MessageSyncResponse',
    'ChatRealtimeJoinResponse', 'CallParticipant', 'RealtimeCall',
    
    # Drive models
    'FileReplica', 'CloudFileObject', 'CloudFile', 'CloudFileIndex',
    'FilePool', 'CloudFolder', 'FileListItem', 'DriveTask', 'UniversalFile',
    'UniversalFileType',
    
    # Live models
    'LiveStream', 'LiveStreamStatus', 'LiveStreamType', 'LiveStreamVisibility',
    
    # Post models
    'Post', 'Publisher', 'PublisherStats', 'PublisherSubscription',
    'ReactInfo', 'PostEmbedView', 'PostEmbedViewRenderer', 'PostAward',
    'PostReaction', 'PostFeaturedRecord', 'PostWithStats', 'PollWithStats',
    'Poll', 'PollQuestion', 'PollOption', 'PollQuestionType', 'PollAnswer',
    'PostTag', 'CategorySubscription', 'PostCategory', 'ScrappedLink',
    'Heatmap', 'HeatmapItem',
    
    # Realm models
    'Realm', 'RealmMember',
    
    # Wallet models
    'Wallet', 'WalletStats', 'WalletPocket', 'Transaction',
    'WalletSubscription', 'WalletSubscriptionRef', 'WalletOrder',
    'WalletGift', 'WalletFund', 'WalletFundRecipient', 'LotteryTicket',
    'LotteryRecord',
]


"""
Drive-related models for Solar Network Python SDK.

This module contains data classes for cloud files, file pools,
folders, and drive operations.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Dict, Any, Union
from enum import Enum


class UniversalFileType(Enum):
    """Universal file types."""
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    FILE = "file"


@dataclass
class UniversalFile:
    """Universal file wrapper."""
    data: Union[CloudFile, Any]
    type: UniversalFileType
    is_link: bool = False
    display_name: Optional[str] = None

    @property
    def is_on_cloud(self) -> bool:
        return isinstance(self.data, CloudFile)

    @property
    def is_on_device(self) -> bool:
        return not self.is_on_cloud


@dataclass
class FileReplica:
    """File replica information."""
    id: str
    object_id: str
    pool_id: str
    pool: Optional[FilePool] = None
    storage_id: str = ""
    status: int = 0
    is_primary: bool = False
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class CloudFileObject:
    """Cloud file object information."""
    id: str
    size: int = 0
    meta: Optional[Dict[str, Any]] = None
    mime_type: Optional[str] = None
    hash: Optional[str] = None
    has_compression: bool = False
    has_thumbnail: bool = False
    file_replicas: List[FileReplica] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class CloudFile:
    """Cloud file information."""
    id: str
    name: str = ""
    description: Optional[str] = None
    file_meta: Optional[Dict[str, Any]] = None
    user_meta: Optional[Dict[str, Any]] = None
    sensitive_marks: List[int] = field(default_factory=list)
    mime_type: Optional[str] = None
    hash: Optional[str] = None
    size: int = 0
    uploaded_at: Optional[datetime] = None
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    url: Optional[str] = None


@dataclass
class CloudFileIndex:
    """Cloud file index information."""
    id: str
    path: str = ""
    file_id: str = ""
    file: Optional[CloudFile] = None
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class FilePool:
    """File pool information."""
    id: str
    name: str = ""
    description: Optional[str] = None
    storage_config: Optional[Dict[str, Any]] = None
    billing_config: Optional[Dict[str, Any]] = None
    policy_config: Optional[Dict[str, Any]] = None
    is_hidden: Optional[bool] = None
    account_id: Optional[str] = None
    resource_identifier: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None


@dataclass
class CloudFolder:
    """Cloud folder information."""
    id: str
    name: str = ""
    parent_folder_id: Optional[str] = None
    account_id: str = ""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)


@dataclass
class FileListItem:
    """File list item."""
    file_index: Optional[CloudFileIndex] = None
    folder_name: Optional[str] = None
    unindexed_file: Optional[CloudFile] = None


@dataclass
class DriveTask:
    """Drive task information."""
    id: str
    task_id: str = ""
    file_name: str = ""
    content_type: str = ""
    file_size: int = 0
    uploaded_bytes: int = 0
    total_chunks: int = 0
    uploaded_chunks: int = 0
    status: int = 0
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    type: str = "FileUpload"  # Task type
    transmission_progress: Optional[float] = None  # Local file upload progress (0.0-1.0)
    error_message: Optional[str] = None
    status_message: Optional[str] = None
    result: Optional[CloudFile] = None
    pool_id: Optional[str] = None
    bundle_id: Optional[str] = None
    encrypt_password: Optional[str] = None
    expired_at: Optional[str] = None

    @property
    def progress(self) -> float:
        """Get upload progress."""
        return self.total_chunks > 0 and self.uploaded_chunks / self.total_chunks or 0.0

    @property
    def estimated_time_remaining(self) -> datetime:
        """Get estimated time remaining."""
        if self.uploaded_bytes == 0 or self.file_size == 0:
            return datetime.now()
        
        remaining_bytes = self.file_size - self.uploaded_bytes
        upload_rate = self.uploaded_bytes / (
            self.created_at - datetime.now()
        ).total_seconds()
        
        if upload_rate == 0:
            return datetime.now()
        
        return datetime.now().replace(
            second=int(remaining_bytes / upload_rate)
        )


"""
Authentication-related models for Solar Network Python SDK.

This module contains data classes for authentication challenges, sessions,
and related security functionality.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Dict, Any


@dataclass
class AppToken:
    """Application token model."""
    token: str


@dataclass
class GeoIpLocation:
    """Geographic IP location information."""
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    country_code: Optional[str] = None
    country: Optional[str] = None
    city: Optional[str] = None


@dataclass
class AuthFactor:
    """Authentication factor information."""
    id: str
    type: int
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    expired_at: Optional[datetime] = None
    enabled_at: Optional[datetime] = None
    trustworthy: int = 0
    created_response: Optional[Dict[str, Any]] = None


@dataclass
class AccountConnection:
    """Account connection information."""
    id: str
    account_id: str
    provider: str
    provided_identifier: str
    meta: Dict[str, Any] = field(default_factory=dict)
    last_used_at: datetime = field(default_factory=datetime.now)
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class AuthChallenge:
    """Authentication challenge information."""
    id: str
    expired_at: Optional[datetime] = None
    step_remain: int = 0
    step_total: int = 0
    failed_attempts: int = 0
    blacklist_factors: List[str] = field(default_factory=list)
    audiences: List[str] = field(default_factory=list)
    scopes: List[str] = field(default_factory=list)
    ip_address: str = ""
    user_agent: str = ""
    nonce: Optional[str] = None
    country_code: Optional[str] = None
    country: Optional[str] = None
    city: Optional[str] = None
    account_id: str = ""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


@dataclass
class AuthSession:
    """Authentication session information."""
    id: str
    label: Optional[str] = None
    last_granted_at: datetime = field(default_factory=datetime.now)
    expired_at: Optional[datetime] = None
    audiences: List[str] = field(default_factory=list)
    scopes: List[str] = field(default_factory=list)
    ip_address: Optional[str] = None
    user_agent: Optional[str] = None
    location: Optional[GeoIpLocation] = None
    type: int = 0
    account_id: str = ""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None


"""
Utility functions for Solar Network Python SDK.

This module provides utility functions for string manipulation,
date formatting, and other common operations.
"""

from typing import Optional


def capitalize_each_word(text: str) -> str:
    """Capitalize the first letter of each word in a string.
    
    Args:
        text: The input string
        
    Returns:
        The string with each word capitalized
    """
    if not text:
        return text
    
    return ' '.join(
        word[0].upper() + word[1:].lower() if word else ''
        for word in text.split(' ')
    )


def format_datetime(dt: Optional[str]) -> Optional[str]:
    """Format a datetime string.
    
    Args:
        dt: The datetime string
        
    Returns:
        The formatted datetime string
    """
    if not dt:
        return None
    
    # Simple formatting - in a real implementation, you might use
    # a library like dateutil or datetime for more robust parsing
    return dt


def get_translation_key(template_key: str) -> str:
    """Get a translation key from a template key.
    
    Args:
        template_key: The template key
        
    Returns:
        The translation key
    """
    parts = template_key.split('_')
    camel_case = ''.join(part.capitalize() for part in parts)
    return f'reaction{camel_case}'


def slugify(text: str) -> str:
    """Convert text to a URL-friendly slug.
    
    Args:
        text: The input text
        
    Returns:
        The slugified text
    """
    import re
    
    # Convert to lowercase and replace spaces with hyphens
    text = text.lower().replace(' ', '-')
    
    # Remove any non-alphanumeric characters except hyphens
    text = re.sub(r'[^a-z0-9\-]', '', text)
    
    # Remove multiple consecutive hyphens
    text = re.sub(r'-+', '-', text)
    
    # Remove leading and trailing hyphens
    text = text.strip('-')
    
    return text


def format_file_size(size_bytes: int) -> str:
    """Format file size in human-readable format.
    
    Args:
        size_bytes: The file size in bytes
        
    Returns:
        The formatted file size
    """
    if size_bytes == 0:
        return "0 B"
    
    size_names = ["B", "KB", "MB", "GB", "TB"]
    import math
    
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    
    return f"{s} {size_names[i]}"


def is_valid_email(email: str) -> bool:
    """Check if an email address is valid.
    
    Args:
        email: The email address to validate
        
    Returns:
        True if the email is valid, False otherwise
    """
    import re
    
    pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    return bool(re.match(pattern, email))


def truncate_string(text: str, max_length: int, suffix: str = "...") -> str:
    """Truncate a string to a maximum length.
    
    Args:
        text: The input text
        max_length: The maximum length
        suffix: The suffix to add when truncating
        
    Returns:
        The truncated string
    """
    if len(text) <= max_length:
        return text
    
    suffix_length = len(suffix)
    if max_length <= suffix_length:
        return suffix[:max_length]
    
    return text[:max_length - suffix_length] + suffix


"""
Setup script for Solar Network Python SDK.

This script sets up the package for distribution.
"""

from setuptools import setup, find_packages
import os

# Read the README file
def read_readme():
    """Read the README file."""
    with open("README.md", "r", encoding="utf-8") as fh:
        return fh.read()

# Read requirements from requirements.txt if it exists
def read_requirements():
    """Read requirements from requirements.txt."""
    requirements_file = "requirements.txt"
    if os.path.exists(requirements_file):
        with open(requirements_file, "r", encoding="utf-8") as fh:
            return [line.strip() for line in fh if line.strip() and not line.startswith("#")]
    return []

setup(
    name="solar-network-sdk",
    version="1.0.0",
    author="liang-work",
    author_email="admin@thsl.dpdns.org",
    description="A Python SDK for interacting with the Solar Network API(Not official)",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/liang-work/SolarNetworkSDKForPython",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Typing :: Typed",
    ],
    python_requires=">=3.8",
    install_requires=read_requirements(),
    extras_require={
        "dev": [
            "pytest>=7.0.0",
            "pytest-asyncio>=0.21.0",
            "black>=22.0.0",
            "flake8>=5.0.0",
            "mypy>=0.991",
        ],
    },
    include_package_data=True,
    package_data={
        "solar_network_sdk": ["py.typed"],
    },
    zip_safe=False,
    keywords="solar network api sdk authentication",
    project_urls={
        "Bug Reports": "https://github.com/liang-work/SolarNetworkSDKForPython/issues",
        "Source": "https://github.com/liang-work/SolarNetworkSDKForPython",
    },
)


"""
Chat-related models for Solar Network Python SDK.

This module contains data classes for chat rooms, messages,
reactions, and real-time communication.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Dict, Any


@dataclass
class ChatRoom:
    """Chat room information."""
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    type: int = 0
    encryption_mode: int = 0
    is_public: bool = False
    is_community: bool = False
    picture: Optional[CloudFile] = None
    background: Optional[CloudFile] = None
    realm_id: Optional[str] = None
    account_id: Optional[str] = None
    realm: Optional[Realm] = None
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    members: Optional[List[ChatMember]] = None
    is_pinned: bool = False


@dataclass
class ChatMessage:
    """Chat message information."""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    id: str = ""
    type: str = "text"
    content: Optional[str] = None
    nonce: Optional[str] = None
    meta: Dict[str, Any] = field(default_factory=dict)
    members_mentioned: List[str] = field(default_factory=list)
    edited_at: Optional[datetime] = None
    attachments: List[CloudFile] = field(default_factory=list)
    reactions: List[ChatReaction] = field(default_factory=list)
    replied_message_id: Optional[str] = None
    forwarded_message_id: Optional[str] = None
    sender_id: str = ""
    sender: Optional[ChatMember] = None
    chat_room_id: str = ""


@dataclass
class ChatReaction:
    """Chat reaction information."""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    id: str = ""
    message_id: str = ""
    sender_id: str = ""
    sender: Optional[ChatMember] = None
    symbol: str = ""
    attitude: int = 0


@dataclass
class ChatMember:
    """Chat member information."""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    id: str = ""
    chat_room_id: str = ""
    chat_room: Optional[ChatRoom] = None
    account_id: str = ""
    account: Optional[Account] = None
    nick: Optional[str] = None
    notify: int = 0
    joined_at: Optional[datetime] = None
    break_until: Optional[datetime] = None
    timeout_until: Optional[datetime] = None
    status: Optional[AccountStatus] = None
    last_typed: Optional[datetime] = None


@dataclass
class ChatSummary:
    """Chat summary information."""
    unread_count: int = 0
    last_message: Optional[ChatMessage] = None


@dataclass
class ChatOnlineAccount:
    """Chat online account information."""
    id: str = ""
    name: str = ""
    nick: str = ""


@dataclass
class ChatOnlineStatus:
    """Chat online status information."""
    online_count: int = 0
    direct_message_status: Optional[AccountStatus] = None
    online_user_names: List[str] = field(default_factory=list)
    online_accounts: List[ChatOnlineAccount] = field(default_factory=list)


@dataclass
class MessageSyncResponse:
    """Message sync response."""
    messages: List[ChatMessage] = field(default_factory=list)
    total_count: int = 0
    current_timestamp: datetime = field(default_factory=datetime.now)


@dataclass
class CallParticipant:
    """Call participant information."""
    identity: str = ""
    name: str = ""
    joined_at: datetime = field(default_factory=datetime.now)


@dataclass
class ChatRealtimeJoinResponse:
    """Chat real-time join response."""
    provider: str = ""
    endpoint: str = ""
    token: str = ""
    call_id: str = ""
    room_name: str = ""
    is_admin: bool = False
    participants: List[CallParticipant] = field(default_factory=list)


@dataclass
class RealtimeCall:
    """Real-time call information."""
    id: str = ""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    deleted_at: Optional[datetime] = None
    ended_at: Optional[datetime] = None
    sender_id: str = ""
    sender: Optional[ChatMember] = None
    room_id: str = ""
    room: Optional[ChatRoom] = None
    upstream_config: Dict[str, Any] = field(default_factory=dict)
    provider_name: Optional[str] = None
    session_id: Optional[str] = None


"""
Pagination utilities for Solar Network Python SDK.

This module provides pagination support for API responses.
"""

from __future__ import annotations
from typing import List, Optional, TypeVar, Generic, Callable, Awaitable
from dataclasses import dataclass, field
from abc import ABC, abstractmethod


T = TypeVar('T')


@dataclass
class PaginationState(Generic[T]):
    """Pagination state information."""
    items: List[T] = field(default_factory=list)
    is_loading: bool = False
    is_reloading: bool = False
    total_count: Optional[int] = None
    has_more: bool = True
    cursor: Optional[str] = None


class PaginationController(ABC, Generic[T]):
    """Abstract pagination controller."""
    
    @property
    @abstractmethod
    def total_count(self) -> Optional[int]:
        """Get the total count of items."""
        pass
    
    @property
    @abstractmethod
    def fetched_count(self) -> int:
        """Get the number of fetched items."""
        pass
    
    @property
    @abstractmethod
    def fetched_all(self) -> bool:
        """Check if all items have been fetched."""
        pass
    
    @property
    @abstractmethod
    def is_loading(self) -> bool:
        """Check if currently loading."""
        pass
    
    @property
    @abstractmethod
    def is_reloading(self) -> bool:
        """Check if currently reloading."""
        pass
    
    @property
    @abstractmethod
    def has_more(self) -> bool:
        """Check if there are more items to fetch."""
        pass
    
    @has_more.setter
    @abstractmethod
    def has_more(self, value: bool) -> None:
        """Set the has_more flag."""
        pass
    
    @property
    @abstractmethod
    def cursor(self) -> Optional[str]:
        """Get the current cursor."""
        pass
    
    @cursor.setter
    @abstractmethod
    def cursor(self, value: Optional[str]) -> None:
        """Set the current cursor."""
        pass
    
    @abstractmethod
    async def fetch(self) -> List[T]:
        """Fetch the next page of items."""
        pass
    
    @abstractmethod
    async def refresh(self) -> None:
        """Refresh the pagination state."""
        pass
    
    @abstractmethod
    async def fetch_further(self) -> None:
        """Fetch more items."""
        pass


class AsyncPaginationController(PaginationController[T]):
    """Asynchronous pagination controller."""
    
    def __init__(self, fetch_func: Callable[[], Awaitable[List[T]]]):
        """Initialize the pagination controller.
        
        Args:
            fetch_func: Function to fetch items
        """
        self._fetch_func = fetch_func
        self._state = PaginationState[T]()
        self._total_count: Optional[int] = None
    
    @property
    def total_count(self) -> Optional[int]:
        """Get the total count of items."""
        return self._total_count
    
    @property
    def fetched_count(self) -> int:
        """Get the number of fetched items."""
        return 0 if self._state.is_reloading else len(self._state.items)
    
    @property
    def fetched_all(self) -> bool:
        """Check if all items have been fetched."""
        return not self._state.has_more or (
            self._total_count is not None and 
            self.fetched_count >= self._total_count
        )
    
    @property
    def is_loading(self) -> bool:
        """Check if currently loading."""
        return self._state.is_loading
    
    @property
    def is_reloading(self) -> bool:
        """Check if currently reloading."""
        return self._state.is_reloading
    
    @property
    def has_more(self) -> bool:
        """Check if there are more items to fetch."""
        return self._state.has_more
    
    @has_more.setter
    def has_more(self, value: bool) -> None:
        """Set the has_more flag."""
        self._state.has_more = value
    
    @property
    def cursor(self) -> Optional[str]:
        """Get the current cursor."""
        return self._state.cursor
    
    @cursor.setter
    def cursor(self, value: Optional[str]) -> None:
        """Set the current cursor."""
        self._state.cursor = value
    
    async def fetch(self) -> List[T]:
        """Fetch the next page of items."""
        return await self._fetch_func()
    
    async def refresh(self) -> None:
        """Refresh the pagination state."""
        self._state = PaginationState(
            items=[],
            is_loading=True,
            is_reloading=True,
            total_count=None,
            has_more=True,
            cursor=None,
        )
        
        new_items = await self.fetch()
        
        self._state = PaginationState(
            items=new_items,
            is_loading=False,
            is_reloading=False,
            total_count=self._total_count,
            has_more=self._state.has_more,
            cursor=self._state.cursor,
        )
    
    async def fetch_further(self) -> None:
        """Fetch more items."""
        if self.fetched_all or self.is_loading:
            return
        
        self._state.is_loading = True
        
        new_items = await self.fetch()
        
        self._state = PaginationState(
            items=self._state.items + new_items,
            is_loading=False,
            is_reloading=self._state.is_reloading,
            total_count=self._total_count,
            has_more=self._state.has_more,
            cursor=self._state.cursor,
        )


class PaginationFiltered(Generic[T]):
    """Pagination with filtering support."""
    
    def __init__(self, controller: PaginationController[T]):
        """Initialize with a pagination controller.
        
        Args:
            controller: The pagination controller
        """
        self._controller = controller
        self._current_filter: Optional[dict] = None
    
    @property
    def current_filter(self) -> Optional[dict]:
        """Get the current filter."""
        return self._current_filter
    
    @current_filter.setter
    def current_filter(self, value: Optional[dict]) -> None:
        """Set the current filter."""
        self._current_filter = value
    
    async def apply_filter(self, filter_params: dict) -> None:
        """Apply a filter and refresh the data.
        
        Args:
            filter_params: The filter parameters
        """
        if self._current_filter == filter_params:
            return
        
        self._state = PaginationState(
            items=[],
            is_loading=True,
            is_reloading=True,
            total_count=None,
            has_more=True,
            cursor=None,
        )
        
        self._current_filter = filter_params
        
        new_items = await self._controller.fetch()
        
        self._state = PaginationState(
            items=new_items,
            is_loading=False,
            is_reloading=False,
            total_count=self._controller.total_count,
            has_more=self._controller.has_more,
            cursor=self._controller.cursor,
        )